/*
 * Java port by Jim Darby. Based on:
 *
 * Copyright (c) 2014 Adafruit Industries
 * Author: Tony DiCola
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package Jimbo.Devices;

import Jimbo.Graphics.Colour;
import Jimbo.Graphics.ColourMatrix;
import Jimbo.Graphics.ColourMatrixDemo;
import Jimbo.Graphics.MatrixHelper;

import java.io.IOException;

import java.util.logging.Logger;
import java.util.logging.Level;

import com.pi4j.Pi4J;
import com.pi4j.exception.Pi4JException;
import com.pi4j.io.spi.Spi;
import com.pi4j.io.spi.SpiProvider;
import com.pi4j.io.gpio.digital.DigitalOutput;
import com.pi4j.io.gpio.digital.DigitalOutputProvider;

public class ST7789 extends MatrixHelper <Colour> implements ColourMatrix
{
    public ST7789 (int width, int height, Spi spi, DigitalOutput dc, DigitalOutput bl) throws com.pi4j.io.exception.IOException
    {
        super (width, height);
        
        this.spi = spi;
        this.dc  = dc;
        this.bl  = bl;
        
        bl.off ();
    }
    
    public void initialise () throws com.pi4j.io.exception.IOException, IOException
    {
	command (ST7789_SWRESET);	// Software reset
        sleep (150);			// delay 150 ms

        command (ST7789_MADCTL);
        data (0x70);

        command (ST7789_FRMCTR2);	// Frame rate ctrl - idle mode
        data (0x0C);
        data (0x0C);
        data (0x00);
        data (0x33);
        data (0x33);

        command (ST7789_COLMOD);
        data (0x05);

        command (ST7789_GCTRL);
        data (0x14);

        command (ST7789_VCOMS);
        data (0x37);

        command (ST7789_LCMCTRL);	// Power control
        data (0x2C);

        command (ST7789_VDVVRHEN);	// Power control
        data (0x01);

        command (ST7789_VRHS);		// Power control
        data (0x12);

        command (ST7789_VDVS);		// Power control
        data (0x20);

        command (0xD0);
        data (0xA4);
        data (0xA1);

        command (ST7789_FRCTRL2);
        data (0x0F);

        command (ST7789_GMCTRP1);	// Set Gamma
        data (0xD0);
        data (0x04);
        data (0x0D);
        data (0x11);
        data (0x13);
        data (0x2B);
        data (0x3F);
        data (0x54);
        data (0x4C);
        data (0x18);
        data (0x0D);
        data (0x0B);
        data (0x1F);
        data (0x23);

        command (ST7789_GMCTRN1);	// Set Gamma
        data (0xD0);
        data (0x04);
        data (0x0C);
        data (0x11);
        data (0x13);
        data (0x2C);
        data (0x3F);
        data (0x44);
        data (0x51);
        data (0x2F);
        data (0x1F);
        data (0x1F);
        data (0x20);
        data (0x23);

        if (invert)
            command (ST7789_INVON);	// Invert display
        else
            command (ST7789_INVOFF);	// Don't invert display

        command (ST7789_SLPOUT);

        command (ST7789_DISPON);	// Display on

        sleep (100);			// 100 ms
        
        bl.on ();
        
        try
        {
            for (int x = 0; x < WIDTH; ++x)
                for (int y = 0; y < HEIGHT; ++y)
                    setPixel (x, y, 255, 0, 0);

            show ();
    
            Thread.sleep (1000);
            
            for (int x = 0; x < WIDTH; ++x)
                for (int y = 0; y < HEIGHT; ++y)
                    setPixel (x, y, 0, 255, 0);

            show ();
    
            Thread.sleep (250);
            
            for (int x = 0; x < WIDTH; ++x)
                for (int y = 0; y < HEIGHT; ++y)
                    setPixel (x, y, 0, 0, 255);

            show ();
    
            Thread.sleep (250);
        }
        
        catch (InterruptedException ex)
        {
            LOGGER.log (Level.SEVERE, "Arse! {0}", ex);
        }
        
    }

    @Override
    public void setPixel (int x, int y, int r, int g, int b)
    {
        if (x < 0 || x >= WIDTH || y < 0 || y >= HEIGHT)
            throw new IllegalArgumentException ("ST7789 Invalid Pixel [" + x + "," + y + "]");
        
        if (r < 0 || r > 255 || g < 0 || g > 255 || b < 0 || b > 255)
           throw new IllegalArgumentException ("ST7789 Invalid Colour (" + r + "," + g + "," + b + ")");
        
        if ((r & 0x04) != 0)
        {
            r += 0x04;
            
            if (r > 255)
                r = 255;
        }
        
        if ((g & 0x02) != 0)
        {
            g += 0x02;
            
            if (g > 255)
                g = 255;
        }
        
        if ((b & 0x04) != 0)
        {
            b += 0x04;
            
            if (b > 255)
                b = 255;
        }
        
        final int value = ((r >> 3) << 11) | ((g >> 2) << 5) | (b >> 3);
        
        // This is really horrible. It will break on non-square displays!
        final int index = ((HEIGHT - 1 - y) + (WIDTH - 1 - x) * WIDTH) * 2;
        
        image[index    ] = (byte) (value >> 8);
        image[index + 1] = (byte) value;
        
        //LOGGER.log (Level.INFO, "Set [{0},{1}] to ({2},{3},{4})", new Object[]{x, y, r, g, b});
    }
    
    public void window (int x0, int y0, int x1, int y1) throws com.pi4j.io.exception.IOException, IOException
    {
        command (ST7789_CASET);         // Column addr set
        data (x0 >> 8);
        data (x0 & 0xFF);               // XSTART
        data (x1 >> 8);
        data (x1 & 0xFF);               // XEND
        command (ST7789_RASET);         // Row addr set
        data (y0 >> 8);
        data (y0 & 0xFF);               // YSTART
        data (y1 >> 8);
        data (y1 & 0xFF);               // YEND
        command (ST7789_RAMWR);         // write to RAM
    }
    
    @Override
    public void show () throws IOException
    {
        //LOGGER.log (Level.INFO, "show");
        
        try
        {
            window (0, 0, WIDTH - 1, HEIGHT - 1);
            data (image);
        }
        
        catch (com.pi4j.io.exception.IOException e)
        {
            throw new IOException ("Got Pi4J IOException " + e.getMessage ());
        }
    }
    
    private void data (int x) throws IOException, com.pi4j.io.exception.IOException
    {
	if (x < 0 || x > 0xff)
	    throw new IllegalArgumentException ("ST7789 bad data value " + x);
        
        dc.on ();
        
        buffer[0] = (byte) x;
        
        spi.transfer (buffer);
    }
    
    private void data (byte[] x) throws IOException, com.pi4j.io.exception.IOException
    {
        int done = 0;
        
        dc.on ();
        
        while (done < x.length)
        {
            final int left = x.length - done;
            final int todo = (left > MAX_TRANSFER) ? MAX_TRANSFER : left;
            
            //LOGGER.log (Level.INFO, "Writing {0} at {1}", new Object[] {todo, done});
            
            spi.transfer (image, done, todo);
            
            done += todo;
        }
    }

    private void command (int x) throws com.pi4j.io.exception.IOException, IOException
    {
	if (x < 0 || x > 0xff)
	    throw new IllegalArgumentException ("ST7789 bad command value " + x);
        
        dc.off ();
        
        buffer[0] = (byte) x;
        
        spi.transfer (buffer);
    }
    
    private void sleep (int ms)
    {
        try
        {
            Thread.sleep (ms);
        }
        
        catch (InterruptedException e)
        {
            LOGGER.log (Level.SEVERE, "ST7789: interrupted sleep: {0}", e.getMessage ());
        }
    }

    private final int MAX_TRANSFER = 4096;
    private final int BITS_PER_PIXEL = 16;
    
    final private boolean invert = true;
    private final Spi spi;
    private final DigitalOutput dc;
    private final DigitalOutput bl;
    private final byte[] buffer = new byte[1];
    private final byte[] image = new byte[WIDTH*HEIGHT*BITS_PER_PIXEL/8];
    
    
    static final private int BG_SPI_CS_BACK = 0;
    static final private int BG_SPI_CS_FRONT = 1;

    static final private int SPI_CLOCK_HZ = 16000000;

    static final private int ST7789_NOP = 0x00;
    static final private int ST7789_SWRESET = 0x01;
    static final private int ST7789_RDDID = 0x04;
    static final private int ST7789_RDDST = 0x09;

    static final private int ST7789_SLPIN = 0x10;
    static final private int ST7789_SLPOUT = 0x11;
    static final private int ST7789_PTLON = 0x12;
    static final private int ST7789_NORON = 0x13;

    static final private int ST7789_INVOFF = 0x20;
    static final private int ST7789_INVON = 0x21;
    static final private int ST7789_DISPOFF = 0x28;
    static final private int ST7789_DISPON = 0x29;

    static final private int ST7789_CASET = 0x2A;
    static final private int ST7789_RASET = 0x2B;
    static final private int ST7789_RAMWR = 0x2C;
    static final private int ST7789_RAMRD = 0x2E;

    static final private int ST7789_PTLAR = 0x30;
    static final private int ST7789_MADCTL = 0x36;
    static final private int ST7789_COLMOD = 0x3A;

    static final private int ST7789_FRMCTR1 = 0xB1;
    static final private int ST7789_FRMCTR2 = 0xB2;
    static final private int ST7789_FRMCTR3 = 0xB3;
    static final private int ST7789_INVCTR = 0xB4;
    static final private int ST7789_DISSET5 = 0xB6;

    static final private int ST7789_GCTRL = 0xB7;
    static final private int ST7789_GTADJ = 0xB8;
    static final private int ST7789_VCOMS = 0xBB;

    static final private int ST7789_LCMCTRL = 0xC0;
    static final private int ST7789_IDSET = 0xC1;
    static final private int ST7789_VDVVRHEN = 0xC2;
    static final private int ST7789_VRHS = 0xC3;
    static final private int ST7789_VDVS = 0xC4;
    static final private int ST7789_VMCTR1 = 0xC5;
    static final private int ST7789_FRCTRL2 = 0xC6;
    static final private int ST7789_CABCCTRL = 0xC7;

    static final private int ST7789_RDID1 = 0xDA;
    static final private int ST7789_RDID2 = 0xDB;
    static final private int ST7789_RDID3 = 0xDC;
    static final private int ST7789_RDID4 = 0xDD;

    static final private int ST7789_GMCTRP1 = 0xE0;
    static final private int ST7789_GMCTRN1 = 0xE1;

    static final private int ST7789_PWCTR6 = 0xFC;
    
    public static void main (String args[]) throws Pi4JException
    {
        LOGGER.fine ("Starting");
        
        var pi4j = Pi4J.newAutoContext ();
        
        // create SPI config
        var config = Spi.newConfigBuilder (pi4j)
                .id ("my-spi-device")
                .name ("My SPI Device")
                .address (BG_SPI_CS_FRONT)
                .baud (SPI_CLOCK_HZ)
                .build ();
        
        LOGGER.fine ("Part #2");
        
         // get a SPI I/O provider from the Pi4J context
        SpiProvider spiProvider = pi4j.provider ("pigpio-spi");
        
        LOGGER.fine ("Part #3");
        
        // use try-with-resources to auto-close SPI when complete
        try (var spi = spiProvider.create(config))
        {
            LOGGER.fine ("Part #4");
            
            final DigitalOutputProvider digitalOutputProvider = pi4j.provider ("pigpio-digital-output");
            final var dc_config = DigitalOutput.newConfigBuilder (pi4j).address (9).build ();
            final var dc = digitalOutputProvider.create (dc_config);
            final var bl_config = DigitalOutput.newConfigBuilder (pi4j).address (19).build ();
            final var bl = digitalOutputProvider.create (bl_config);
            
            ST7789 dev = new ST7789 (240, 240, spi, dc, bl);
            
            dev.initialise ();
            
            ColourMatrixDemo demo = new ColourMatrixDemo (dev);
            
            demo.run ();
        }

        catch (Exception e)
        {
            LOGGER.log (Level.SEVERE, "Oops! {0}", e.getMessage ());
        }
        
        LOGGER.fine ("Closing down");

        pi4j.shutdown ();
        
        LOGGER.fine ("And we're done…");
    }
    
    /** Logging. */
    private static final Logger LOGGER = Logger.getLogger (ST7789.class.getName ());
}
